#pragma once

// C++ system level
#include <string>
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"

// project specific
#include "LambdaTask.hpp"
#include "SpiderLib/Callback.hpp"
#include "SpiderLib/CallbackMap.hpp"

// qthing stuff
// #include ""



namespace SiliconTorch {

  namespace CyanBus {

    // CyanBus payload MTU
    // This leads to ~54 P/s @ 2 MBaud at 100% load
    constexpr uint32_t MTU = 4096;

    // CyanBus envelope header
    extern const char* const HEADER;


    typedef struct {
      const uint8_t* payload;
      const uint16_t length;
    } PacketData;

    typedef struct {
      const bool crcOK;
      const uint16_t payloadLEN;
    } PacketMetrics;


    class CyanBus {
      public:
        CyanBus(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate = 115200, uint8_t uartChannel = 1);

        uint32_t getBaudRate();
        void setBaudRate(uint32_t baudRate);

        const SpiderLib::CallbackMap<void(const PacketData&)> packetCallback;
        const SpiderLib::Callback<void(const PacketMetrics&)> metricsCallback;

      private:

        QueueHandle_t uartEvQ;
        QueueHandle_t packetQ;

        uint8_t uartChannel;
        uint8_t tx, rx, de, re;

        void txEN(bool state);
        void rxEN(bool state);
        void txrxEN(bool state);


        // Receive only ATM!  (make public after sound implementation…)
        void write(std::string& data);
        void write(const uint8_t* data, uint32_t length);


        void readerTask();
        void packetTask();

        Util::LambdaTask* readerTaskHandle = NULL;
        Util::LambdaTask* packetTaskHandle = NULL;

    };
  }
}
